from django.core.cache import caches
from django.http import Http404, HttpResponse
from django_filters.rest_framework import DjangoFilterBackend
from office365.runtime.client_request_exception import ClientRequestException
from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.exceptions import ErrorDetail, PermissionDenied
from rest_framework.filters import OrderingFilter, SearchFilter

from unicef_sharepoint.client import SharePointClient, SharePointClientException
from unicef_sharepoint.filters import SharePointLibraryFilter
from unicef_sharepoint.lib import File
from unicef_sharepoint.models import SharePointLibrary, SharePointSite, SharePointTenant
from unicef_sharepoint.serializers import (
    SharePointFileSerializer,
    SharePointItemSerializer,
    SharePointLibrarySerializer,
    SharePointSiteSerializer,
    SharePointTenantSerializer,
)
from unicef_sharepoint.utils import get_cache_key

cache = caches['default']


class AbstractSharePointViewSet(viewsets.ReadOnlyModelViewSet):
    serializer_class = None
    filter_backends = (SearchFilter, DjangoFilterBackend, OrderingFilter)

    def get_library(self):
        return SharePointLibrary.objects.get(
            site__tenant__url__contains=self.tenant, name=self.folder, site__name=self.site)

    @property
    def client(self):
        key = self.get_cache_key(**{'client': 'client'})
        client = cache.get(key)
        if client is None:
            dl = self.get_library()
            dl_info = {
                'url': dl.site.site_url(),
                'relative_url': dl.site.relative_url(),
                'folder': dl.name
            }
            if dl.site.tenant.username:
                dl_info['username'] = dl.site.tenant.username
                dl_info['password'] = dl.site.tenant.password
            try:
                client = SharePointClient(**dl_info)
                cache.set(key, client)
            except SharePointClientException:
                raise PermissionDenied

        return client

    def get_serializer_context(self):
        ctx = super().get_serializer_context()
        ctx.update({
            'tenant': self.tenant,
            'site': self.site,
            'folder': self.folder
        })
        return ctx

    def handle_exception(self, exc):
        response = super().handle_exception(exc)
        if isinstance(exc, Http404):
            response.data['detail'] = ErrorDetail('No document found using selected filters', 'not_found')
        return response

    @property
    def tenant(self):
        return self.kwargs.get('tenant')

    @property
    def site(self):
        return self.kwargs.get('site')

    @property
    def folder(self):
        return self.kwargs.get('folder')

    def get_cache_key(self, **kwargs):
        key = get_cache_key([self.tenant, self.site, self.folder], **kwargs)
        return key


class ItemSharePointViewSet(AbstractSharePointViewSet):
    serializer_class = SharePointItemSerializer

    def get_queryset(self):
        kwargs = self.request.query_params.dict()
        try:
            key = self.get_cache_key(**kwargs)
            response = cache.get(key)
            if response is None:
                response = self.client.read_items(filters=kwargs)
                cache.set(key, response)
            return response
        except ClientRequestException:
            raise Http404


class ItemSharePointCamlViewSet(ItemSharePointViewSet):
    def get_queryset(self):
        kwargs = self.request.query_params.dict()
        cache_dict = kwargs.copy()
        cache_dict['caml'] = 'true'
        try:
            key = self.get_cache_key(**cache_dict)
            response = cache.get(key)
            if response is None:
                response = self.client.read_caml_items(filters=kwargs)
                cache.set(key, response)
            return response
        except ClientRequestException:
            raise Http404


class FileSharePointViewSet(AbstractSharePointViewSet):
    serializer_class = SharePointFileSerializer
    lookup_field = 'filename'

    def get_object(self):
        filename = self.kwargs.get('filename', None)
        try:
            filename, *extension = filename.split('__ext__')
            if extension and len(extension) == 1:
                extension = extension[0]
            else:
                extension = 'pdf'
            doc_file = self.client.read_file(f'{filename}.{extension}')
        except ClientRequestException:
            raise Http404
        return doc_file

    def get_queryset(self):
        kwargs = self.request.query_params.dict()
        try:
            return self.client.read_files(filters=kwargs)
        except ClientRequestException:
            raise Http404

    @action(detail=True, methods=['get'])
    def download(self, request, *args, **kwargs):
        sh_file = self.get_object()
        relative_url = sh_file.properties['ServerRelativeUrl']
        response = File.open_binary(self.client.context, relative_url)

        django_response = HttpResponse(
            content=response.content,
            status=response.status_code,
            content_type=response.headers['Content-Type'],
        )
        django_response['Content-Disposition'] = 'attachment; filename=%s' % sh_file.properties['Name']
        return django_response


class SharePointTenantViewSet(viewsets.ReadOnlyModelViewSet):
    filter_backends = (SearchFilter, DjangoFilterBackend, OrderingFilter)
    queryset = SharePointTenant.objects.all()
    serializer_class = SharePointTenantSerializer
    search_fields = ('url', )


class SharePointSiteViewSet(viewsets.ReadOnlyModelViewSet):
    filter_backends = (SearchFilter, DjangoFilterBackend, OrderingFilter)
    queryset = SharePointSite.objects.all()
    serializer_class = SharePointSiteSerializer
    search_fields = ('name', )


class SharePointLibraryViewSet(viewsets.ReadOnlyModelViewSet):
    filter_backends = (SearchFilter, DjangoFilterBackend, OrderingFilter)
    queryset = SharePointLibrary.objects.all()
    serializer_class = SharePointLibrarySerializer
    search_fields = ('name', )
    filterset_class = SharePointLibraryFilter



import logging
import os

from django.conf import settings
from office365.runtime.auth.authentication_context import AuthenticationContext
from office365.sharepoint.client_context import ClientContext
from office365.sharepoint.file_creation_information import FileCreationInformation
from office365.sharepoint.helpers.querystring_builder import QueryStringBuilder

from unicef_sharepoint.lib import CamlQueryBuilder, File

logger = logging.getLogger(__name__)


class SharePointClientException(BaseException):
    """SharePoint Exception when initializing the client"""


class SharePointClient:
    """Client to access SharePoint Document Library"""

    def __init__(self, *args, **kwargs) -> None:
        self.site_path = kwargs.get('url', settings.SHAREPOINT_TENANT['url'])
        self.relative_url = kwargs.get('relative_url', None)
        username = kwargs.get('username', settings.SHAREPOINT_TENANT['user_credentials']['username'])
        password = kwargs.get('password', settings.SHAREPOINT_TENANT['user_credentials']['password'])
        self.folder = kwargs.get('folder', 'Documents')
        ctx_auth = AuthenticationContext(url=self.site_path)
        if ctx_auth.acquire_token_for_user(username=username, password=password):
            self.context = ClientContext(self.site_path, ctx_auth)
        else:
            logger.exception(ctx_auth.get_last_error())
            raise SharePointClientException(ctx_auth.get_last_error())

    def get_folder(self, list_title):
        list_obj = self.context.web.lists.get_by_title(list_title)
        folder = list_obj.root_folder
        self.context.load(folder)
        self.context.execute_query()
        logger.info(f'List url: {folder.properties["ServerRelativeUrl"]}')
        return folder

    def read_folders(self, list_title):
        self.get_folder(list_title)
        folders = self.context.web.folders
        self.context.load(folders)
        self.context.execute_query()
        for folder in folders:
            logger.info(f'Folder name: {folder.properties["Name"]}')

        return folders

    def read_files(self, filters=dict()):
        querystring = QueryStringBuilder(filters).get_querystring()
        folder = self.get_folder(self.folder)
        files = folder.files.filter(querystring)
        self.context.load(files)
        self.context.execute_query()
        for cur_file in files:
            logger.info(f'File name: {cur_file.properties["Name"]}')

        return files

    def read_items(self, filters=dict()):
        querystring = QueryStringBuilder(filters).get_querystring()
        list_object = self.context.web.lists.get_by_title(self.folder)
        items = list_object.get_items().filter(querystring)
        self.context.load(items)
        self.context.execute_query()
        return items

    def read_file(self, filename):
        folder = self.get_folder(self.folder)
        cur_file = folder.files.get_by_url(f'/{self.relative_url}/{self.folder}/{filename}')
        self.context.load(cur_file)
        self.context.execute_query()
        logger.info(f'File name: {cur_file.properties["Name"]}')
        return cur_file

    def read_caml_items(self, filters=dict(), scope=None):
        """Read a folder example"""
        list_obj = self.context.web.lists.get_by_title(self.folder)
        qry = CamlQueryBuilder(filters, scope).get_query()
        items = list_obj.get_items(qry)
        self.context.execute_query()
        return items

    def upload_file_alt(self, target_folder, name, content):
        context = target_folder.context
        info = FileCreationInformation()
        info.content = content
        info.url = name
        info.overwrite = True
        target_file = target_folder.files.add(info)
        context.execute_query()
        return target_file

    def upload_file(self, path, list_title='Documents', upload_into_library=True):
        with open(path, 'rb') as content_file:
            file_content = content_file.read()

        if upload_into_library:
            target_folder = self.context.web.lists.get_by_title(list_title).root_folder
            file = self.upload_file_alt(target_folder, os.path.basename(path), file_content)
            logger.info('File url: {}'.format(file.properties['ServerRelativeUrl']))
        else:
            target_url = f'/{self.folder}/{os.path.basename(path)}'
            File.save_binary(self.context, target_url, file_content)

    def download_file(self, filename):
        response = File.open_binary(self.context, f'/{self.folder}/{filename}')
        with open(f'./data/{filename}', 'wb') as local_file:
            local_file.write(response.content)



from django.contrib import admin

from .models import SharePointLibrary, SharePointSite, SharePointTenant


@admin.register(SharePointTenant)
class SharepointTenantAdmin(admin.ModelAdmin):
    search_fields = ('url', )
    list_display = ('url', )


@admin.register(SharePointSite)
class SharepointSiteAdmin(admin.ModelAdmin):
    search_fields = ('name', )
    list_display = ('name', 'site_type', 'tenant')
    list_filter = ('site_type', 'tenant')


@admin.register(SharePointLibrary)
class DocumentLibraryAdmin(admin.ModelAdmin):
    search_fields = ('name', )
    list_display = ('name', 'site', 'active', 'public')
    list_filter = ('active', 'public')



# Generated by Django 3.0.3 on 2020-02-07 19:25

import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='SharePointTenant',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('url', models.URLField(unique=True)),
                ('username', models.CharField(blank=True, max_length=64, null=True, verbose_name='Username')),
                ('password', models.CharField(blank=True, max_length=64, null=True, verbose_name='Password')),
            ],
            options={
                'verbose_name': 'SharePoint Tenant',
                'verbose_name_plural': 'SharePoint Tenants',
                'ordering': ['url'],
            },
        ),
        migrations.CreateModel(
            name='SharePointSite',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('name', models.CharField(max_length=32, verbose_name='Name')),
                ('site_type', models.CharField(choices=[('sites', 'sites'), ('teams', 'teams')], default='sites', max_length=16, verbose_name='Site Type')),
                ('tenant', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='sites', to='unicef_sharepoint.SharePointTenant')),
            ],
            options={
                'verbose_name': 'SharePoint Site',
                'verbose_name_plural': 'SharePoint Sites',
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='SharePointLibrary',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('name', models.CharField(max_length=64, verbose_name='Name')),
                ('active', models.BooleanField(default=True, verbose_name='Active')),
                ('public', models.BooleanField(default=True, verbose_name='Public')),
                ('site', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='libraries', to='unicef_sharepoint.SharePointSite')),
            ],
            options={
                'verbose_name': 'SharePoint Document Library',
                'verbose_name_plural': 'SharePoint Document Libraries',
                'ordering': ['name'],
                'unique_together': {('name', 'site')},
            },
        ),
    ]



from django.conf import settings
from office365.sharepoint.helpers.utils import to_camel
from rest_framework import serializers
from rest_framework.reverse import reverse

from unicef_sharepoint.models import SharePointLibrary, SharePointSite, SharePointTenant


class SharePointPropertyField(serializers.ReadOnlyField):
    """tries to get attribute from the object or properties"""

    def get_attribute(self, instance):
        camel_case = to_camel(self.source)
        if getattr(instance, 'properties') and camel_case in instance.properties:
            return instance.properties[camel_case]
        return super().get_attribute(instance)


class SharePointPropertyManyField(serializers.ReadOnlyField):
    """tries to get attribute from the object or properties"""

    def get_attribute(self, instance):
        camel_case = to_camel(self.source)
        if getattr(instance, 'properties') and camel_case in instance.properties:
            values = instance.properties[camel_case]
            if values:
                values = values.replace('; ', ';').split(';')
            return values
        return super().get_attribute(instance)


class UpperSharePointPropertyField(serializers.ReadOnlyField):

    def get_attribute(self, instance):
        upper_case = self.source.upper()
        if getattr(instance, 'properties') and upper_case in instance.properties:
            return instance.properties[upper_case]
        return super().get_attribute(instance)


class SharePointTenantSerializer(serializers.ModelSerializer):
    class Meta:
        model = SharePointTenant
        exclude = ('username', 'password')


class SharePointSiteSerializer(serializers.ModelSerializer):
    class Meta:
        model = SharePointSite
        fields = '__all__'


class SharePointLibrarySerializer(serializers.ModelSerializer):
    site_name = serializers.ReadOnlyField(source='site.name')
    api_url = serializers.SerializerMethodField()

    def get_api_url(self, obj):
        reverse_url = reverse('unicef_sharepoint:sharepoint-list',
                              kwargs={'tenant': obj.site.tenant.name, 'site': obj.site.name, 'folder': obj.name})
        return settings.HOST + reverse_url

    class Meta:
        model = SharePointLibrary
        fields = ('name', 'site_name', 'active', 'library_url', 'api_url')


class SharePointItemSerializer(serializers.Serializer):

    id = UpperSharePointPropertyField()
    guid = UpperSharePointPropertyField()
    created = SharePointPropertyField()
    modified = SharePointPropertyField()
    title = SharePointPropertyField()
    url = SharePointPropertyField()
    resource_url = serializers.ReadOnlyField()
    download_url = serializers.SerializerMethodField()

    def get_download_url(self, obj):
        filename = obj.properties.get('Title', '')
        if filename:
            k = filename.rfind(".")
            filename = filename[:k] + "__ext__" + filename[k + 1:]
        relative_url = reverse('unicef_sharepoint:sharepoint-files-download', kwargs={
            'tenant': self.context['tenant'],
            'site': self.context['site'],
            'folder': self.context['folder'],
            'filename': filename
        })
        return f'{settings.HOST}{relative_url}'


class SharePointFileSerializer(serializers.Serializer):
    name = SharePointPropertyField()
    type_name = serializers.ReadOnlyField()
    url = serializers.ReadOnlyField()
    linking_uri = SharePointPropertyField()
    server_relative_url = SharePointPropertyField()
    unique_id = SharePointPropertyField()
    title = SharePointPropertyField()
    time_created = SharePointPropertyField()
    time_last_modified = SharePointPropertyField()

    def get_download_url(self, obj):
        relative_url = reverse('unicef_sharepoint:sharepoint-files-download', kwargs={
            'tenant': self.context['tenant'],
            'site': self.context['site'],
            'folder': self.context['folder'],
            'filename': obj.properties['Name'].split('.')[0]})
        return f'{settings.HOST}{relative_url}'



from rest_framework import routers

from unicef_sharepoint.views import (
    FileSharePointViewSet,
    ItemSharePointCamlViewSet,
    ItemSharePointViewSet,
    SharePointLibraryViewSet,
    SharePointSiteViewSet,
    SharePointTenantViewSet,
)

app_name = 'unicef_sharepoint'

router = routers.DefaultRouter()

router.register(r'tenants', SharePointTenantViewSet, basename='sharepoint-tenant')
router.register(r'sites', SharePointSiteViewSet, basename='sharepoint-site')
router.register(r'libraries', SharePointLibraryViewSet, basename='sharepoint-library')
router.register(r'sharepoint/(?P<tenant>[\w\-]+)/(?P<site>[\w\-]+)/(?P<folder>[\w|\W]+)/files',
                FileSharePointViewSet, basename='sharepoint-files')
router.register(r'sharepoint/(?P<tenant>[\w\-]+)/(?P<site>[\w\-]+)/(?P<folder>[\w|\W]+)',
                ItemSharePointViewSet, basename='sharepoint')
router.register(r'sharepoint-caml/(?P<tenant>[\w\-]+)/(?P<site>[\w\-]+)/(?P<folder>[\w|\W]+)',
                ItemSharePointCamlViewSet, basename='sharepoint-caml')

urlpatterns = router.urls



import collections


def get_cache_key(args, **kwargs):
    joint_args = ''.join(args)
    order_dict = collections.OrderedDict(kwargs)
    return hash(joint_args) + hash(frozenset(order_dict.items()))



from urllib.parse import quote

from django.db import models
from django.utils.translation import gettext as _
from model_utils.models import TimeStampedModel


class SharePointTenant(TimeStampedModel):
    url = models.URLField(unique=True)
    username = models.CharField(verbose_name=_("Username"), max_length=64, null=True, blank=True)
    password = models.CharField(verbose_name=_("Password"), max_length=64, null=True, blank=True)

    @property
    def name(self):
        return self.url.split('//')[-1].split('.')[0]

    def __str__(self):
        return f'{self.url}'

    class Meta:
        ordering = ['url']
        verbose_name = 'SharePoint Tenant'
        verbose_name_plural = 'SharePoint Tenants'


class SharePointSite(TimeStampedModel):
    SITE = 'sites'
    TEAM = 'teams'

    SITE_TYPES = (
        (SITE, SITE),
        (TEAM, TEAM),
    )

    tenant = models.ForeignKey(SharePointTenant, related_name='sites', on_delete=models.deletion.CASCADE)
    name = models.CharField(verbose_name=_("Name"), max_length=32)
    site_type = models.CharField(verbose_name=_("Site Type"), max_length=16, choices=SITE_TYPES, default=SITE)

    class Meta:
        ordering = ['name']
        verbose_name = 'SharePoint Site'
        verbose_name_plural = 'SharePoint Sites'

    def __str__(self):
        return f'{self.tenant} ({self.name})'

    def relative_url(self):
        return f'{self.site_type}/{self.name}'

    def site_url(self):
        return f'{self.tenant}{self.site_type}/{self.name}'


class SharePointLibrary(TimeStampedModel):
    name = models.CharField(verbose_name=_("Name"), max_length=64)
    site = models.ForeignKey(SharePointSite, related_name='libraries', on_delete=models.deletion.CASCADE)
    active = models.BooleanField(verbose_name=_("Active"), default=True)
    public = models.BooleanField(verbose_name=_("Public"), default=True)

    class Meta:
        ordering = ['name']
        unique_together = ('name', 'site', )
        verbose_name = 'SharePoint Document Library'
        verbose_name_plural = 'SharePoint Document Libraries'

    def __str__(self):
        return f'{self.name} ({self.site.name}) [{self.site.tenant}]'

    @property
    def library_url(self):
        return str(self.site.tenant) + quote(f'{self.site.site_type}/{self.site.name}/{self.name}')



#!/usr/bin/env python
import ast
import os
import re

from setuptools import find_packages, setup

HERE = os.path.abspath(os.path.dirname(__file__))
init = os.path.join(HERE, "src", "unicef_sharepoint", "__init__.py")

_version_re = re.compile(r'__version__\s+=\s+(.*)')
_name_re = re.compile(r'NAME\s+=\s+(.*)')

with open(init, 'rb') as f:
    content = f.read().decode('utf-8')
    VERSION = str(ast.literal_eval(_version_re.search(content).group(1)))
    NAME = str(ast.literal_eval(_name_re.search(content).group(1)))


setup(
    name=NAME,
    version=VERSION,
    url='https://github.com/unicef/unicef-sharepoint',
    author='UNICEF',
    author_email='rapidpro@unicef.org',
    description='Provides Basic Unicef User model and integration with Azure',
    platforms=['any'],
    license='Apache 2 License',
    classifiers=[
        'Environment :: Web Environment',
        'Programming Language :: Python :: 3.8',
        'Framework :: Django',
        'Framework :: Django :: 3.0',
        'Intended Audience :: Developers'],
    install_requires = [
        'django',
        'django-filter',
        'django-model-utils',
        'djangorestframework',
        'office365-rest-python-client',
    ],
    extras_require={
        'test': [
            'django-webtest',
            'factory-boy',
            'flake8',
            'isort',
            'mock',
            'pytest',
            'pytest-cov',
            'pytest-django',
            'pytest-echo',
            'pytest-pythonpath',
            'pytest-redis',
            'requests-mock',
            'unittest2',
            'vcrpy',
            'drf_api_checker',
        ],
    },
    package_dir={'': 'src'},
    packages=find_packages('src'),
    include_package_data=True,
)



from django_filters import rest_framework as filters

from unicef_sharepoint.models import SharePointLibrary


class SharePointLibraryFilter(filters.FilterSet):

    class Meta:
        model = SharePointLibrary
        fields = {
            'site': ['exact', 'in'],
            'active': ['exact', ],
        }



import logging

from office365.runtime.utilities.http_method import HttpMethod
from office365.runtime.utilities.request_options import RequestOptions
from office365.sharepoint.caml_query import CamlQuery
from office365.sharepoint.file import File
from office365.sharepoint.helpers.utils import to_camel

logger = logging.getLogger(__name__)


def recursive_builder(queries, operator='And'):
    if queries:
        query = queries.pop()
        if len(queries) == 0:
            return query
        elif len(queries) == 1:
            last_query = queries.pop()
            return f'<{operator}>' + query + last_query + f'</{operator}>'
        else:
            return f'<{operator}>' + query + recursive_builder(queries) + f'</{operator}>'
    return ''


class CamlQueryBuilder:
    date_operators = ['Geq', 'Gt', 'Leq', 'Lt']
    mapping_operator = {
        'gte': 'Geq',
        'gt': 'Gt',
        'lte': 'Leq',
        'lt': 'Lt',
        'not': 'Neq',
        'contains': 'Contains',
        'eq': 'Eq',
    }

    filters = {}

    def __init__(self, filters, scope=None):
        super().__init__()
        self.scope = scope

        if filters:
            self.filters = filters

    def create_query(self):
        where_condition = ''

        if self.filters.keys():
            filter_queries = []
            for filter_name, filter_value in self.filters.items():
                querystring_operator = filter_name.split('__')[-1]
                operator = self.mapping_operator.get(querystring_operator, 'Eq')

                filter_name = to_camel(filter_name.split('__')[0])
                if operator in self.date_operators:
                    column_type, value = 'DateTime', "{}T00:00:00Z".format(filter_value)  # 2016-03-26
                    query = '<{}><FieldRef Name="{}" /><Value Type="{}">{}</Value></{}>'.format(
                        operator, filter_name, column_type, value, operator)
                elif operator == 'Contains':
                    column_type, values = 'Note', filter_value
                    query = '<{}><FieldRef Name="{}" /><Value Type="{}">{}</Value></{}>'.format(
                            operator, filter_name, column_type, filter_value, operator)
                else:
                    column_type, values = 'Text', filter_value.split(',')
                    query = ''
                    for value in values:
                        query = query + '<{}><FieldRef Name="{}" /><Value Type="{}">{}</Value></{}>'.format(
                            operator, filter_name, column_type, value, operator)
                    if len(values) > 1:
                        query = f'<Or>{query}</Or>'
                filter_queries.append(query)
            where_condition = recursive_builder(filter_queries)
            if len(filter_queries) > 1:
                where_condition = f"<And>{where_condition}</And>"

        scope = f' Scope="{self.scope}"' if self.scope else ''
        query = f'<View{scope}><Query><Where>{where_condition}</Where></Query></View>'
        return query

    def get_query(self):
        return CamlQuery.create_custom_query(self.create_query())


class File(File):
    @staticmethod
    def open_binary(ctx, server_relative_url):
        url = r"{0}web/getfilebyserverrelativeurl('{1}')/\$value".format(ctx.serviceRootUrl, server_relative_url)
        request = RequestOptions(url)
        request.method = HttpMethod.Get
        response = ctx.execute_request_direct(request)
        return response
